;; Generated package description from engrave-faces.el  -*- no-byte-compile: t -*-
(define-package "engrave-faces" "0.3.1" "Convert font-lock faces to other formats" '((emacs "27.1")) :commit "fe29d9b593a0f7ada4df3f52dfb9f7f8c1bdaaa7" :authors '(("TEC" . "https://github/tecosaur")) :maintainer '("TEC" . "tec@tecosaur.com") :keywords '("faces") :url "https://github.com/tecosaur/engrave-faces")
